// DualNode.h

#pragma once // for visual studio.
#ifndef DUALNODE_H // for other compilers.
#define DUALNODE_H

/* recommended. */
#include "Node.h"
/*/
// use this if something wrong with code before.
#ifndef NODE_H
#include "Node.h"
#endif // end for #ifndef NODE_H.
//*/

template<typename T>
//  ()    T  2     DualNode<T> *.
//    .
struct DualNode : /*public virtual DataContainer<T>,*/ protected Node<T>
{
private:
	typedef DualNode<T> Dual; // make code below smaller.

	//    .
	Dual *prev;

public:
	using Node<T>::GetData;
	using Node<T>::SetData;

public:
	//      .
	virtual Dual *const GetNext() const { return static_cast<Dual *>(Node<T>::GetNext()); }
	//      next,
	//       next' 
	//   (this),  next   nullptr.
	virtual void SetNext(Dual *const &next)
	{
		Dual *tNext = Dual::GetNext();
		if(tNext == next) return;
		Dual *nPrev = (next ? next->prev : nullptr);
		Node<T>::SetNext(next);
		if(this == nPrev) return;
		if(next != nullptr) next->Dual::SetPrev(this);
		if(tNext != nullptr && this == tNext->prev) tNext->Dual::SetPrev(nPrev);
		if(nPrev != nullptr) nPrev->Dual::SetNext(tNext);
	}
	//virtual void SetNext(Dual &next) { SetNext(&next); }

	//      .
	virtual Dual *const GetPrev() const { return prev; }
	//      prev,
	//       prev' 
	//   (this),  prev   nullptr.
	virtual void SetPrev(Dual *const &prev)
	{
		Dual *tPrev = this->prev;
		if(tPrev == prev) return;
		Dual *pNext = (prev ? prev->Dual::GetNext() : nullptr);
		this->prev = prev;
		if(this == pNext) return;
		if(prev != nullptr) prev->Dual::SetNext(this);
		if(tPrev != nullptr && this == tPrev->Dual::GetNext()) tPrev->Dual::SetNext(pNext);
		if(pNext != nullptr) pNext->Dual::SetPrev(tPrev);
	}
	//virtual void SetPrev(Dual &prev) { Dual::SetPrev(&prev); }
	//  SetNext(Dual *)  SetPrev(Dual *)   ,
	//     ,   
	//        8   .

	/* use this code if compiler support C++11 or higher */

	DualNode() : Node<T>(), prev(nullptr) { }
	DualNode(T const &data) : Node<T>(data), prev(nullptr) { }
	DualNode(T const &data, Dual *const &next)
		: Node<T>(data), prev(nullptr) { Dual::SetNext(next); }
	//DualNode(T const &data, Dual &next) : Dual(data, &next) { }
	DualNode(T const &data, Dual *const &next, Dual *const &prev)
		: Node<T>(data), prev(nullptr) { Dual::SetNext(next); Dual::SetPrev(prev); }
	//DualNode(T const &data, Dual &next, Dual &prev) : Dual(data, &next, &prev) { }

	// Copy constructor.
	DualNode(Dual const &lvalue) : Node<T>(lvalue.Dual::GetData(),
		lvalue.Dual::GetNext() == &lvalue ? this : nullptr),
		prev(lvalue.prev == &lvalue ? this : nullptr) { }

	DualNode(T &&data) : Node<T>(std::move<T &>(data)), prev(nullptr) { }
	DualNode(T &&data, Dual *const &next)
		: Node<T>(std::move<T &>(data)), prev(nullptr)
	{
		Dual::SetNext(next);
	}
	DualNode(T &&data, Dual *const &next, Dual *const &prev)
		: Node<T>(std::move<T &>(data)), prev(nullptr)
	{
		Dual::SetNext(next);
		Dual::SetPrev(prev);
	}
	// Move constructor.
	DualNode(Dual &&rvalue) : Node<T>(std::move<Dual &>(rvalue)),
		prev(rvalue.prev == &rvalue ? this : rvalue.prev)
	{
		rvalue.prev = nullptr;
		Dual *next = Dual::GetNext();
		if(next) next->prev = this;
		if(prev) prev->Node<T>::SetNext(this);
	}
	/*/
	// use code below if code before generate error.

	DualNode() : prev(nullptr) { Node<T>(); }
	DualNode(T const &data) : prev(nullptr) { Node<T>(data); }
	DualNode(T const &data, Dual *const next) : prev(nullptr)
	{
		Node<T>(data);
		Dual::SetNext(next);
	}
	DualNode(T const &data, Dual &next) : prev(nullptr)
	{
		Node<T>(data);
		Dual::SetNext(&next);
	}
	DualNode(T const &data, Dual *const next, Dual *const prev) : prev(nullptr)
	{
		Node<T>(data);
		Dual::SetNext(next);
		Dual::SetPrev(prev);
	}
	DualNode(T const &data, Dual &next, Dual &prev) : prev(nullptr)
	{
		Node<T>(data);
		Dual::SetNext(&next);
		Dual::SetPrev(&prev);
	}
	DualNode(Dual const &lvalue) : prev(nullptr) {
		Node<T>(lvalue.Dual::GetData(),
			lvalue.Dual::GetNext() == &lvalue ? this : nullptr),
			prev(lvalue.prev == &lvalue ? this : nullptr);
	}
	//*/

	virtual ~DualNode()
	{
		if(prev) Dual::SetNext(this);
		prev = nullptr;
	}

	// Copy assignment operator.
	virtual Dual const &operator=(Dual const &lvalue)
	{
		Node<T>::operator=(lvalue);
		Node<T>::SetNext(lvalue.Dual::GetNext() == &lvalue ? this : nullptr);
		prev = (lvalue.prev == &lvalue ? this : nullptr);
		return *this;
	}
	// Move assignment operator.
	virtual Dual const &operator=(Dual &&rvalue)
	{
		Node<T>::operator=(std::move<Dual &>(rvalue));
		prev = (rvalue.prev == &rvalue ? this : rvalue.prev);
		rvalue.prev = nullptr;
		Dual *next = Dual::GetNext();
		if(next) next->prev = this;
		if(prev) prev->Node<T>::SetNext(this);
		return *this;
	}

protected:
	//      next,
	//       next' 
	//   (this),  next   nullptr.
	virtual void SetNext(Node<T> *const &next) { SetNext(dynamic_cast<Dual *>(next)); }
};
#endif // end for #ifndef DUALNODE_H.
