// Node.h

#pragma once // for visual studio.
#ifndef NODE_H // for other compilers.
#define NODE_H

#pragma warning(disable : 4355) // for visual studio only.

template<typename T>
//  ()    T  1     Node<T> *.
struct Node //: public virtual DataContainer<T>
{
private:
	// ,     .
	T data;
	//    .
	Node<T> *next;
	//        ,  
	//      ,
	//   (볺,     )
	//       .

public:
	//  .
	virtual T const &GetData() const { return data; }
	//  .
	virtual void SetData(T const &data) { this->data = data; }
	//  .
	virtual void SetData(T &&data) { this->data = std::move<T &>(data); }

	//      .
	virtual Node<T> *const GetNext() const { return next; }
	//      next.
	virtual void SetNext(Node<T> *const &next) { this->next = next; }
	//void SetNext(Node<T> &next) { SetNext(&next); }
	// ,       .
	// _GetNext()  _SetNext(Node<T> *)      Next.

	Node() : data(/*T()*/), next(nullptr) { }
	Node(T const &data) : data(data), next(nullptr) { }
	Node(T const &data, Node<T> *const &next) : data(data), next(next) { }
	//Node(T const &data, Node<T> &next) : data(data), next(&next) { }

	// Copy constructor.
	Node(Node<T> const &lvalue) : data(lvalue.data), next(lvalue.next) { }

	Node(T &&data) : data(std::move<T &>(data)), next(nullptr) { }
	Node(T &&data, Node<T> *const &next) : data(std::move<T &>(data)), next(next) { }
	// Move constructor.
	Node(Node<T> &&rvalue) : data(std::move<T &>(rvalue.data)),
		next(rvalue.next == &rvalue ? this : rvalue.next) { rvalue.next = nullptr; }
	// VS2008 does not support rvalue references (&&).

	virtual ~Node() { next = nullptr; }

	// Copy assignment operator.
	virtual Node<T> const &operator=(Node<T> const &lvalue)
	{
		data = lvalue.data;
		next = lvalue.next;
		return *this;
	}
	// Move assignment operator.
	virtual Node<T> const &operator=(Node<T> &&rvalue)
	{
		data = std::move<T &>(rvalue.data);
		next = (rvalue.next == &rvalue ? this : rvalue.next);
		rvalue.next = nullptr;
		return *this;
	}
};
#endif // end for #ifndef NODE_H.
